@@ -3,6 +3,7 @@ import amqp
 import traceback
 import logging
 import threading
+import multiprocessing
 import queue
 import time
 
@@ -28,7 +29,8 @@ class Connector:
 					ssl                    = None,
 					poll_rate              = 0.25,
 					prefetch               = 1,
-					prefetch_size          = 0,
+					session_fetch_limit    = None,
+					durable                = False,
 				):
 
 		# The synchronous flag controls whether the connector should limit itself
@@ -38,7 +40,15 @@ class Connector:
 		self.synchronous   = synchronous
 		self.poll_rate     = poll_rate
 		self.prefetch      = prefetch
-		self.prefetch_size = prefetch_size
+		self.durable       = durable
+
+		# The session fetch limit allows control of the total number of
+		# AMQP messages the instance of `Connector()` will *ever* fetch in it's
+		# entire lifetime.
+		# If none, there will be no limit.
+		self.session_fetch_limit = session_fetch_limit
+		self.session_fetched     = 0
+		self.queue_fetched       = 0
 
 		# Number of tasks that have been retreived by this client.
 		# Used for limiting the number of tasks each client will pre-download and
@@ -50,6 +60,7 @@ class Connector:
 		self.log.info("Setting up AqmpConnector!")
 
 
+		self.log.info("Fetch limit: '%s'", self.session_fetch_limit)
 		self.log.info("Comsuming from queue '{conq}', emitting responses on '{tasq}'.".format(conq = task_queue, tasq=response_queue))
 		self.master = master
 
@@ -106,8 +117,11 @@ class Connector:
 
 
 		# set up the task and response queues.
-		self.taskQueue = queue.Queue()
-		self.responseQueue = queue.Queue()
+		# These need to be multiprocessing queues because
+		# messages can sometimes be inserted from a different process
+		# then the interface is created in.
+		self.taskQueue = multiprocessing.Queue()
+		self.responseQueue = multiprocessing.Queue()
 
 
 		# Threading logic
@@ -131,34 +145,34 @@ class Connector:
 		# Channel and exchange setup
 		self.channel = self.connection.channel()
 		self.channel.basic_qos(
-				prefetch_size  = self.prefetch_size,
+				prefetch_size  = 0,
 				prefetch_count = self.prefetch,
 				a_global       = False
 			)
 
 	def _setupQueues(self):
 
-		self.channel.exchange_declare(self.task_exchange, type=self.task_exchange_type, auto_delete=False)
-		self.channel.exchange_declare(self.response_exchange, type=self.response_exchange_type, auto_delete=False)
+		self.channel.exchange_declare(self.task_exchange,     type=self.task_exchange_type,     auto_delete=False, durable=self.durable)
+		self.channel.exchange_declare(self.response_exchange, type=self.response_exchange_type, auto_delete=False, durable=self.durable)
 
 		# set up consumer and response queues
 		if self.master:
 			# Master has to declare the response queue so it can listen for responses
-			self.channel.queue_declare(self.response_q, auto_delete=False)
-			self.channel.queue_bind(self.response_q, exchange=self.response_exchange, routing_key=self.response_q.split(".")[0])
+			self.channel.queue_declare(self.response_q, auto_delete=False, durable=self.durable)
+			self.channel.queue_bind(   self.response_q, exchange=self.response_exchange, routing_key=self.response_q.split(".")[0])
 			self.log.info("Binding queue {queue} to exchange {ex}.".format(queue=self.response_q, ex=self.response_exchange))
 
 		if not self.master:
 			# Clients need to declare their task queues, so the master can publish into them.
-			self.channel.queue_declare(self.task_q, auto_delete=False)
-			self.channel.queue_bind(self.task_q, exchange=self.task_exchange, routing_key=self.task_q.split(".")[0])
+			self.channel.queue_declare(self.task_q, auto_delete=False, durable=self.durable)
+			self.channel.queue_bind(   self.task_q, exchange=self.task_exchange, routing_key=self.task_q.split(".")[0])
 			self.log.info("Binding queue {queue} to exchange {ex}.".format(queue=self.task_q, ex=self.task_exchange))
 
 		# "NAK" queue, used for keeping the event loop ticking when we
 		# purposefully do not want to receive messages
 		# THIS IS A SHITTY WORKAROUND for keepalive issues.
-		self.channel.queue_declare('nak.q', auto_delete=False)
-		self.channel.queue_bind('nak.q', exchange=self.response_exchange, routing_key="nak")
+		self.channel.queue_declare('nak.q', auto_delete=False, durable=self.durable)
+		self.channel.queue_bind('nak.q',    exchange=self.response_exchange, routing_key="nak")
 
 
 	def _poll_proxy(self):
@@ -169,6 +183,7 @@ class Connector:
 			self.log.warning("AQMP Connector thread interrupted by keyboard interrupt!")
 			self._poll()
 
+
 	def _poll(self):
 		'''
 		Internal function.
@@ -184,7 +199,9 @@ class Connector:
 		integrator = 0               # Time since last status message emitted.
 		loop_delay = self.poll_rate  # Poll interval for queues.
 
-		while self.run:
+		# When run is false, don't halt until
+		# we've flushed the outgoing items out the queue
+		while self.run or self.responseQueue.qsize():
 
 			try:
 				# Kick over heartbeat
@@ -196,7 +213,7 @@ class Connector:
 				self.connection.heartbeat_tick()
 				self.connection.send_heartbeat()
 				time.sleep(loop_delay)
-				if self.active == 0 or not self.synchronous:
+				if (self.active == 0 or not self.synchronous) and self.run:
 
 					if integrator > print_time:
 						self.log.info("Looping, waiting for job.")
@@ -207,7 +224,6 @@ class Connector:
 						self.log.info("Active task running.")
 
 				self._publishOutgoing()
-
 				# Reset the print integrator.
 				if integrator > 5:
 					integrator = 0
@@ -242,6 +258,8 @@ class Connector:
 			# Prevent never breaking from the loop if the feeding queue is backed up.
 			if ret > self.prefetch:
 				break
+			if self.atFetchLimit():
+				break
 
 			item = self.channel.basic_get(queue=in_queue)
 			if item:
@@ -249,14 +267,18 @@ class Connector:
 				item.channel.basic_ack(item.delivery_info['delivery_tag'])
 				self.taskQueue.put(item.body)
 				ret += 1
+
+				self.session_fetched += 1
+				if self.atFetchLimit():
+					self.log.info("Session fetch limit reached. Not fetching any additional content.")
 			else:
 				break
 
-		self.log.info("Retreived %s items!", ret)
+		if ret:
+			self.log.info("Retreived %s items!", ret)
 		return ret
 
 	def _publishOutgoing(self):
-
 			if self.master:
 				out_queue = self.task_exchange
 				out_key   = self.task_q.split(".")[0]
@@ -267,8 +289,10 @@ class Connector:
 			while 1:
 				try:
 					put = self.responseQueue.get_nowait()
-					self.log.info("Publishing message of len '%0.3f'K to exchange '%s'", len(put)/1024, out_queue)
+					# self.log.info("Publishing message of len '%0.3f'K to exchange '%s'", len(put)/1024, out_queue)
 					message = amqp.basic_message.Message(body=put)
+					if self.durable:
+						message.properties["delivery_mode"] = 2
 					self.channel.basic_publish(message, exchange=out_queue, routing_key=out_key)
 					self.active -= 1
 
@@ -276,22 +300,58 @@ class Connector:
 					break
 
 
+
+	def atFetchLimit(self):
+		'''
+		Track the fetch-limit for the active session. Used to allow an instance to connect,
+		fetch one (and only one) item, and then do things with the fetched item without
+		having the background thread fetch and queue a bunch more items while it's working.
+		'''
+		if not self.session_fetch_limit:
+			return False
+
+		return self.session_fetched >= self.session_fetch_limit
+
+	def atQueueLimit(self):
+		'''
+		Track the fetch-limit for the active session. Used to allow an instance to connect,
+		fetch one (and only one) item, and then do things with the fetched item without
+		having the background thread fetch and queue a bunch more items while it's working.
+		'''
+		if not self.session_fetch_limit:
+			return False
+
+		return self.queue_fetched >= self.session_fetch_limit
+
+
 	def getMessage(self):
 		'''
 		Try to fetch a message from the receiving Queue.
 		Returns the method if there is one, False if there is not.
 		Non-Blocking.
 		'''
+
+		if self.atQueueLimit():
+			raise ValueError("Out of fetchable items!")
+
 		try:
 			put = self.taskQueue.get_nowait()
+			self.queue_fetched += 1
 			return put
 		except queue.Empty:
 			return None
 
-	def putMessage(self, message):
+	def putMessage(self, message, synchronous=False):
 		'''
 		Place a message into the outgoing queue.
+
+		if synchronous is true, this call will block until
+		the items in the outgoing queue are less then the
+		value of synchronous
 		'''
+		if synchronous:
+			while self.responseQueue.qsize() > synchronous:
+				time.sleep(0.1)
 		self.responseQueue.put(message)
 
 
@@ -303,6 +363,11 @@ class Connector:
 		'''
 		self.log.info("Stopping AMQP interface thread.")
 		self.run = False
+		while self.responseQueue.qsize() > 0:
+			self.log.info("%s remaining outgoing AMQP items.", self.responseQueue.qsize())
+			time.sleep(1)
+
+
 		self.thread.join()
 		self.log.info("AMQP interface thread halted.")
 
